
`include "fifo_rtl.sv"
`include "cache_mem.sv"
module cache_controller(
                        //user interface
                        output logic data_vld, busy,  
                        input  logic rd, wr,
                        input logic  [MAIN_MEM_ADDRESS_WIDTH-1:0] addr,                     
                        output logic [MEM_DATA_WIDTH-1:0]         rd_data, 
                        input logic  [MEM_DATA_WIDTH-1:0]         wr_data,  
                        // main memory interface 
                        output logic mem_rd, mem_wr, 
                        output logic [MAIN_MEM_ADDRESS_WIDTH-1:0] mem_addr, 
                        output logic [MEM_DATA_WIDTH-1:0]         mem_wr_data,  
                        input logic  [MEM_DATA_WIDTH-1:0]         mem_rd_data, 
                        // Signals needed by checker                        
                        output logic cache_exists,  // exists at specified address
                        output logic [MEM_DATA_WIDTH-1:0]         cache_data,

                        // clock and resets 
                        input logic clk, 
                        input logic rst_n);
  
  // Local declarations
  timeunit 1ns;     timeprecision 100ps;
    logic [MAIN_MEM_ADDRESS_WIDTH-1:0] wr_addr;  // address to write into cache 
  // ***** TO PUT IN A checker  chk_rd_wr_cntlr 
   property p_wr_rd;
    int v_addr, v_data;
      first_match(($rose(wr), v_addr=addr, v_data=wr_data) ##1 
      !(wr && addr==v_addr)[*1:$] ##1 rd && addr==v_addr) |->
                                   ##[1:MEM_CYCLES] rd_data==v_data;
  endproperty : p_wr_rd
  ap_wr_rd : assert property(@ (posedge clk) p_wr_rd);

     mp_no_simultaneous_rd_wr: assume property (
     @ (posedge clk) not (rd && wr));

    mp_busy_no_rd_wr: assume property (
            @ (posedge clk) busy |-> !rd && !wr);
  // *****
  // ***** chk_rd_cntrl
  property p_read_miss;
	  logic [MAIN_MEM_ADDRESS_WIDTH-1:0] v_addr;
 	  (rd && miss, v_addr=addr) |-> ##MEM_CYCLES data_vld ##1 cache_mem1.cache_valid[v_addr];
  endproperty : p_read_miss
  ap_read_miss : assert property(p_read_miss);

   function logic[CACHE_SIZE-1 :0] get_index(logic [MAIN_MEM_ADDRESS_WIDTH-1:0] address,
     logic [CACHE_SIZE-1:0]    cache_entry, 
     logic [(CACHE_SIZE-1):0] [MAIN_MEM_ADDRESS_WIDTH-1 : 0] cache_mem_addr);
	//logic success=0;
	for (int j=0; j<= CACHE_SIZE-1; j=j+1) begin : for0
        // $display("infinite loop here ???  L65");             
        if(cache_entry[j]==1 && cache_mem_addr[j]==address) begin : if1// found line
//          found_existing_entry = 1'b1;
          get_index = j;
          //success =1;
		  // $display("L68 wr_cache existing no invalidate, found_index=%d", found_index);
		  break;
          // $display("found existing entry in wr L70");
        end : if1 
    end : for0
  endfunction : get_index

  property p_write2cache_exits;
	  logic [MAIN_MEM_ADDRESS_WIDTH-1:0] v_addr;
	  logic [MEM_DATA_WIDTH-1:0] 		 v_wr_data;
      logic [MAIN_MEM_ADDRESS_WIDTH-1:0]  v_index;
  	   (wr && cache_exists, v_addr=addr, v_wr_data=wr_data) |=> 
          (wr_cache, v_index=get_index(
             v_addr, cache_mem1.cache_entry, cache_mem1.cache_mem_addr)) ##0
			cache_mem1.cache_valid[v_addr] && 
            cache_mem1.cache_mem_addr[v_index] == v_addr ##1 
            cache_mem1.cache_mem_data[v_index] == v_wr_data;
  endproperty : p_write2cache_exits
  ap_write2cache_exits : assert property(p_write2cache_exits);

  // cover rd and miss
  cp_rd_miss: cover property(rd && miss);

  // cover rd and hit
  cp_rd_hit: cover property(rd && cache_exists);

   // Verify that if a Cache miss, then the correct data is fetched  and outputted 
	  property p_miss_then_data_2output;
		logic [MEM_DATA_WIDTH-1:0] 		 v_data;
		logic [MAIN_MEM_ADDRESS_WIDTH-1:0] v_addr;
        (rd && miss, v_addr=addr) |-> 
            ##MEM_CYCLES (data_vld && mem_addr==v_addr, v_data=mem_rd_data) 
                 |-> rd_data==v_data;
	  endproperty : p_miss_then_data_2output
		ap_miss_then_data_2output : assert property(
              p_miss_then_data_2output);

  // Verify the push and pop operations
	  ap_fifo_pop_on_miss_full: assert property(rd && miss && full |=> pop);
      ap_fifo_push_on_miss: assert property(rd && miss |=> push);

   // --------------------------------------------------		  		
	// Verify that for a user interface read and miss,
    //  busy is asserted for 3 cycles, and at the 3rd cycle data_vld==1
		 ap_read_miss_busy_data_vld: assert property(
				 (rd && miss) |=> busy[*MEM_CYCLES] intersect 
                                   (!data_vld[*MEM_CYCLES-1] ##1 data_vld));

   // if read and hit, thatn rd_data== cache_data
    ap_read_hit : assert property(
       rd && cache_exists |=> data_vld && rd_data==cache_data);

  // ***** chk_invalidate
  function logic check_cache_entry4MT(logic [MAIN_MEM_ADDRESS_WIDTH-1:0] addr2invalidate,
     logic [CACHE_SIZE-1:0]    cache_entry, 
     logic [(CACHE_SIZE-1):0] [MAIN_MEM_ADDRESS_WIDTH-1 : 0] cache_mem_addr);
    automatic logic success   =0; // If==1 then it was not invalidated
    automatic int   i;
    // CVC
    //while (!success || i <= CACHE_SIZE -1) begin : while1
    for (i=0; i <= CACHE_SIZE -1; i++) begin : for1
      if(cache_entry[i]==1 && cache_mem_addr[i] == addr2invalidate) begin 
        success  =1; // found cache line
		    break;
	    end 
    end : for1
    if (success) check_cache_entry4MT  =0; // was not invalidated
    else check_cache_entry4MT	 =1; // was invalidated
  endfunction : check_cache_entry4MT  
  property p_invalidate;
    logic [MAIN_MEM_ADDRESS_WIDTH-1:0] v_addr2invalidate; // address to invalidate
    (invalidate, v_addr2invalidate=addr2invalidate) |=>  cache_mem1.cache_valid[v_addr2invalidate] ==0 
            && check_cache_entry4MT(
                addr2invalidate, cache_mem1.cache_entry, cache_mem1.cache_mem_addr); 
  endproperty : p_invalidate
  ap_invalidate : assert property(p_invalidate);
	
  //  internal  controller signals
  logic[2:0]   wr_cycle_timer;  // write cycle timer 
  logic[2:0]   rd_cycle_timer;  // read cycle timer
  logic [MAIN_MEM_ADDRESS_WIDTH-1:0] addr_r1, addr_r2, addr_r3; 
  logic                              wr_cache;  // instructs cache to write a line 
  logic                              invalidate; // instructs cache to invalidate a line 
//  logic [MAIN_MEM_ADDRESS_WIDTH-1:0] wr_addr;  // address to write into cache 
  logic [MAIN_MEM_ADDRESS_WIDTH-1:0] addr2invalidate; // address to invalidate
  logic [MEM_DATA_WIDTH-1:0]         cache_wr_data; // data to write to cache  
  logic                              check_cache;  // check the cache for exists
  logic                              data_from_mem;
  
  logic [MAIN_MEM_ADDRESS_WIDTH-1:0] addr_r; // delayed address from user interface
  
  typedef   logic [MAIN_MEM_ADDRESS_WIDTH-1 : 0] word_t;
  // typedef   word_t [0 : CACHE_SIZE-1] buffer_t; 
  // registered wr_data for write to mamin mem
  // logic [MEM_DATA_WIDTH-1:0]                      wr_data_r;  // << delete, unused 

  
  logic                                          miss;
  // invalidate, addr_invld;
  
  logic empty, almost_empty, almost_full, full, push, pop;
  word_t last_entry; // last entry of addr into the fifo. Used when cache
  // reaches maximum size
  logic error; // fifo
//  typedef enum {IDLE, WR_MEM1, WR_MEM2, WR_MEM3} mem_write_phase_e; 
	
//   mem_write_phase_e main_mem_wr_phase;
//   typedef enum {IDLE1, RD_MEM1, RD_MEM2, RD_MEM3} mem_read_phase_e;
//   mem_read_phase_e main_mem_rd_phase;

 default clocking default_clk @ (posedge clk); endclocking
  //default disable iff !rst_n;
  // int cache[*]; // associative array (AA) <<-- CHANGE YYPE to word_t ???????
  // assign cache_exists  = cache.exists(addr); 
  // assign cache_data    = cache[addr];
  assign miss            = !cache_exists;
  assign check_cache = rd || wr;

    
  // 1) invalidate cache at popped address if wr and full and cache does not exists 
  // 2) push addr
  
  // 3) if wr and exists, supply data
  // 2. Cache: Cache accesses are very fast, minimally 1 clock cycle. 
  // Cache size is parameterized.  This cache is a write-through cache, 
  // thus the write to memory always proceeds.  Cache is lines are either updated or created on writes.
  
  // 3. User interface write: If a write from user interface occurs, 
  // then write must be forwarded to main memory and to the cache.
  
  
  // **********  FILLING THE CACHE ON WRITES & INVALIDATE IF MISS AND FULL  **********
  // If there is a write cache miss, meaning that there was a cache entry, but we're doing a write, 
  // then the cache line must be updated.  
  // Note: In the FIFO (fifo_rtl.sv), the "data_out" (connected to "last_entry")  always 
  // supplies the last entry. Thus, prior to the pop being executed, the last line is available 
  // on signal last_entry.  
  always @ (posedge clk) begin : filling_cache_invalidates
    addr_r              <= addr; // The addr_r is stored into the fifo
    // defaults assignments 
    pop <= 1'b0;
    push <= 1'b0;
    invalidate <= 1'b0;
	//wr_cache <= 1'b0;
	if((rd || wr) && miss) begin : read_write_miss
//	if((rd || wr) && miss && !busy) begin : read_write_miss	  
	  push    <= 1'b1;  // push the new entry (i.e., the new addr)
	  if(full) begin : rd_wr_full
         pop     <= 1'b1;  // pop the last entry
		 addr2invalidate       <= last_entry;
         invalidate            <= 1'b1;
	  end : rd_wr_full
	  // else if (wr && cache_exists) begin : wr_cache_exists
		  
	 end :read_write_miss
  /*    
    chk_invalidate chk_invalidate_wr_miss_full
      (.cache_mem1.cache_valid(valid),
       .cache_mem1.cache_mem_addr(cache_mem_addr),
       .cache_mem1.cache_entry(cache_entry),
       .addr2check(addr), // addr2invalidate,
       .clk(clk),
       .reset_n(rst_n));
*/
          // chk_NO_invalidate chk_NO_invalidate_wr_miss_exists(.cache_mem1.cache_valid(valid),
          //                .cache_mem1.cache_mem_addr(cache_mem_addr),
          //                .cache_mem1.cache_entry(cache_entry),
          //                .addr2check(addr),
          //                .clk(clk),
          //                .reset_n(rst_n));  
		// chk_invalidate  chk_invalidate_rd_miss(.cache_mem1.cache_valid(valid),
        //                  .cache_mem1.cache_mem_addr(cache_mem_addr),
        //                  .cache_mem1.cache_entry(cache_entry),
        //                  .addr2invalidate(addr2invalidate),
        //                  .clk(clk),
        //                  .reset_n(rst_n));
 
  end : filling_cache_invalidates

  
  // ********** MAIN MEMORY WRITES CONTROL **********
  // Main memory is slow, and requires 3 cycles to read or to write.  
  // Main memory size is parameterized.  
  // Data is latched on the falling edge of the write signal and is driven 
  // on the rising edge of the read signal.  
  // Handle the main write cycles 
    always @ (posedge clk) begin : main_mem_accesses 
      // defaults
      mem_rd   <= 1'b0;
      busy     <= 1'b0;
      data_vld <= 1'b0;
      mem_rd   <= 1'b0;
      mem_wr   <= 1'b0;
	  wr_cache <= 1'b0;
      // WRITES 
      if (wr) begin : wr_2main // Start of a write
//      if (wr && !busy) begin : wr_2main // Start of a write 		
        ap_wr_cycle_timer_not_0: assert (wr_cycle_timer==0) else $error("wr signal when memory counter !=0");
		// ?? 
		ap_wr_cycle_timer_not_0_concurrent: assert property(
                    wr_cycle_timer==0) else $error("wr signal when memory counter !=0");
		
        wr_cycle_timer <= MEM_CYCLES-1'b1; // load cycle timer for latency 
        mem_wr_data       <= wr_data;
        mem_addr          <= addr;
        mem_wr            <= 1'b1;
        busy              <= 1'b1;
        // cache[mem_addr]    = wr_data; // fill the cache
		wr_cache <= 1'b1;
        cache_wr_data <= wr_data;
        wr_addr <= addr; 
      end : wr_2main
      if (wr_cycle_timer!=0) begin : dec_wr_cycle_timer 
        wr_cycle_timer <= wr_cycle_timer - 1'b1;
        mem_wr            <= 1'b1;
        busy              <= 1'b1;
      end  : dec_wr_cycle_timer
      // READS 
      if(rd && !miss) begin : rd_from_cache // must provide the data_vld
//      if(rd && !miss && !busy) begin : rd_from_cache // must provide the data_vld		
        data_vld           <= 1'b1;
      end : rd_from_cache // bagath miss 
      if(rd && miss) begin : rd_from_main
//      if(rd && miss && !busy) begin : rd_from_main		
        ap_rd_cycle_timer_not_zero: assert (rd_cycle_timer==0) 
                   else $error("wr signal when memory counter !=0");
        rd_cycle_timer <= MEM_CYCLES - 1'b1; // load cycle timer for latency 
        mem_addr          <= addr;
        mem_rd            <= 1'b1;
        busy              <= 1'b1;
      end : rd_from_main
      if(rd_cycle_timer!=0) begin : rd_mem_in_progress
        rd_cycle_timer <= rd_cycle_timer - 1'b1; 
        mem_rd            <= 1'b1;
        busy              <= 1'b1;
      end : rd_mem_in_progress
      if(rd_cycle_timer==1) begin : write_to_cache 
        // cache[mem_addr]    = mem_rd_data;
		wr_cache <= 1'b1;
        cache_wr_data <= mem_rd_data;
        wr_addr <= mem_addr; // addr; 
        data_vld          <= 1'b1;    //bagath
        // rd_cycle_timer <= rd_cycle_timer - 1'b1; 
      end : write_to_cache
      if (!rst_n) begin : resets
        rd_cycle_timer <= 0;
        wr_cycle_timer <= 0;
        mem_rd   <= 1'b0;
        busy     <= 1'b0;
        data_vld <= 1'b0;
        mem_rd   <= 1'b0;
        mem_wr   <= 1'b0;
		wr_cache <= 1'b0;
        //cache.delete; // Clears all elements of associative array       
      end : resets      
    end : main_mem_accesses  
  
  

  /********** Read data to user interface **********/
  // ERROR HERE
  // CANNOT rely on the miss signal as the switch
  // Need the delayed signal 
  // assign rd_data = miss? mem_rd_data :  // If miss, return memory data
  //                 cache_data;    // If !miss, return cache data
  always @ (posedge clk) begin : rd_data_output
	if (rd && miss) begin : rd_miss_data_from_mem
//	if (rd && miss && !busy) begin : rd_miss_data_from_mem	
      data_from_mem <= 1'b1;
	end :rd_miss_data_from_mem
	else if (rd && cache_exists) begin : rd_hit
//	else if (rd && cache_exists && !busy) begin : rd_hit	  
      data_from_mem <= 1'b0;	
    end : rd_hit
	  end : rd_data_output

  assign rd_data = data_from_mem && data_vld ? mem_rd_data :  // If miss, return memory data
                  cache_data; // cache[addr];   // If hit, return cache data		



   cache_mem cache_mem1(
     .wr_cache(wr_cache),  // instructs cache to write a line 
     .invalidate(invalidate), // instructs cache to invalidate a line 
     .addr(addr),  // address to determine exists 
     .wr_addr(wr_addr),  // address to write into cache 
     .addr2invalidate(addr2invalidate), // address to invalidate
     .cache_wr_data(cache_wr_data), // data to write to cache
	 .check_cache(check_cache), // check for exists 
     .cache_data(cache_data),  // data read from cache  
     .cache_exists(cache_exists),               // cache exists at rd_addr
     .clk(clk), .reset_n(rst_n));
  
  
 /**********  FIFO, EMULATES LEAST RECENTLY USED ENTRY **********/
  // Stores the address for later invalitae when full 
  fifo_mem LRU_fifo(.clk(clk), 
                    .rst_n(rst_n), 
                    .empty(empty), 
                    .full(full),
                    .data_out(last_entry),
                    .error(error),
                    .data_in(addr_r),
                    .push(push), 
                    .pop(pop));
  
endmodule 
























